class Directories:
    def __init__(self, files, dependencies):
        self.dependencies = dependencies

        # Include directories
        self.include_private = files["include_directories"]
        self.include_public = files["include_directories_public"]

        # Default include path
        # if self.root_directory.joinpath('include').exists():
        #    self._include_directories_public = [self.root_directory.joinpath('include')] + self._include_directories_public

        # Public include directories of dependencies are forwarded
        for target in self.dependencies:
            self.include_public += target.directories.include_public

        # Make unique and resolve
        self.include_private = list(
            dict.fromkeys(dir.resolve() for dir in self.include_private)
        )
        self.include_public = list(
            dict.fromkeys(dir.resolve() for dir in self.include_public)
        )

    def final_directories_list(self):
        return list(dict.fromkeys(
            self.include_private + self.include_public
        ))

    def include_command(self):
        include_directories_command = []
        for directory in self.final_directories_list():
            include_directories_command += ["-I", str(directory)]

        return include_directories_command

    def make_private_directories_public(self):
        self.include_public = self.final_directories_list()
        self.include_private = []


"""Module for the Circle class."""

class Circle(list):
    """List with a circular dependency representation.

    Utility class to print circular dependencies. Given
    a list ["A", "B", "A"], this list will print as:

        A -> B -> A

    indicating the circular dependency.
    """

    def __repr__(self) -> str:
        """Return a representation of this circle.

        Returns
        -------
        str
            An arrow connected string of the circular
            dependency

        """
        return " -> ".join(repr(item) for item in self)

    def __str__(self) -> str:
        """Return a string representation of this circle.

        Returns
        -------
        str
            An arrow connected string of the circular
            dependency

        """
        return " -> ".join(str(item) for item in self)



import os, sys
import unittest
import subprocess
import shutil
import logging
import io
import stat
from pathlib import Path as _Path
from multiprocessing import freeze_support

from clang_build import clang_build
from clang_build import platform
from clang_build.errors import CompileError
from clang_build.errors import LinkError
from clang_build.logging_tools import TqdmHandler as TqdmHandler

def on_rm_error( func, path, exc_info):
    # path contains the path of the file that couldn't be removed
    # let's just assume that it's read-only and try to unlink it.
    try:
        os.chmod( path, stat.S_IWRITE )
        os.unlink( path )
    except:
        print(f'Error trying to clean up file "{path}":\n{exc_info}')

def clang_build_try_except( args ):
    try:
        clang_build.build(clang_build.parse_args(args))
    except CompileError as compile_error:
        logger = logging.getLogger('clang_build')
        logger.error('Compilation was unsuccessful:')
        for target, errors in compile_error.error_dict.items():
            printout = f'Target [{target}] did not compile. Errors:\n'
            printout += ' '.join(errors)
            logger.error(printout)
    except LinkError as link_error:
        logger = logging.getLogger('clang_build')
        logger.error('Linking was unsuccessful:')
        for target, errors in link_error.error_dict.items():
            printout = f'Target [{target}] did not link. Errors:\n{errors}'
            logger.error(printout)

class TestClangBuild(unittest.TestCase):
    def test_hello_world_mwe(self):
        clang_build_try_except(['-d', 'test/mwe'])

        try:
            output = subprocess.check_output(['./build/default/bin/main'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        self.assertEqual(output, 'Hello!')

    def test_build_types(self):
        for build_type in ['release', 'relwithdebinfo', 'debug', 'coverage']:
            clang_build_try_except(['-d', 'test/mwe', '-b', build_type])

            try:
                output = subprocess.check_output([f'./build/{build_type}/bin/main'], stderr=subprocess.STDOUT).decode('utf-8').strip()
            except subprocess.CalledProcessError as e:
                self.fail(f'Could not run compiled program with build type "{build_type}". Message:\n{e.output}')

            self.assertEqual(output, 'Hello!')

    def test_compile_error(self):
        with self.assertRaises(CompileError):
            clang_build.build(clang_build.parse_args(['-d', 'test/build_errors/compile_error', '-V']))

    def test_link_error(self):
        with self.assertRaises(LinkError):
            clang_build.build(clang_build.parse_args(['-d', 'test/build_errors/link_error', '-V']))

    def test_script_call(self):
        try:
            subprocess.check_output(['clang-build', '-d', 'test/mwe', '-V'], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self.fail('Compilation failed')
        try:
            output = subprocess.check_output(['./build/default/bin/main'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        self.assertEqual(output, 'Hello!')

    def test_hello_world_rebuild(self):
        clang_build_try_except(['-d', 'test/mwe', '-V'])

        try:
            output = subprocess.check_output(['./build/default/bin/main'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')
        self.assertEqual(output, 'Hello!')

        ### TODO: the following does not seem to work under coverage runs...
        # logger = logging.getLogger('clang_build')

        # stream_capture = io.StringIO()
        # ch = logging.StreamHandler(stream_capture)
        # ch.setLevel(logging.DEBUG)
        # logger.addHandler(ch)
        # clang_build_try_except(['-d', 'test/mwe', '-V'])
        # logger.removeHandler(ch)
        # self.assertRegex(stream_capture.getvalue(), r'.*\[main\]: target is already compiled*')

        # stream_capture = io.StringIO()
        # ch = logging.StreamHandler(stream_capture)
        # ch.setLevel(logging.DEBUG)
        # logger.addHandler(ch)
        # clang_build_try_except(['-d', 'test/mwe', '-V', '-f'])
        # logger.removeHandler(ch)
        # self.assertRegex(stream_capture.getvalue(), r'.*\[main\]: target needs to build sources*')


    def test_automatic_include_folders(self):
        clang_build_try_except(['-d', 'test/mwe_with_default_folders', '-V'])

        try:
            output = subprocess.check_output(['./build/default/bin/main'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        self.assertEqual(output, 'Calculated Magic: 30')

    def test_toml_mwe(self):
        clang_build_try_except(['-d', 'test/toml_mwe'])

        try:
            output = subprocess.check_output(['./build/default/bin/runHello'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        self.assertEqual(output, 'Hello!')

    def test_toml_custom_folder(self):
        clang_build_try_except(['-d', 'test/toml_with_custom_folder'])

        try:
            output = subprocess.check_output(['./build/default/bin/runHello'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        self.assertEqual(output, 'Hello!')

    def test_pyapi_directory(self):
        clang_build_try_except(['-d', 'test/py-api/directory', '-V'])

        try:
            output = subprocess.check_output(['./build/default/bin/main'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        self.assertEqual(output, 'the version is 1.2.0')

    def test_subproject(self):
        clang_build_try_except(['-d', 'test/subproject', '-V'])

        try:
            output = subprocess.check_output(['./build/myexe/default/bin/runLib'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        self.assertEqual(output, 'Hello! mylib::triple(3) returned 9')

    def test_pyapi_subproject(self):
        clang_build_try_except(['-d', 'test/py-api/subproject', '-V'])

        try:
            output = subprocess.check_output(['./build/myexe/default/bin/runLib'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        self.assertEqual(output, 'Hello! mylib::triple(3) returned 9')

    def test_boost_filesystem(self):
        clang_build_try_except(['-d', 'test/boost-filesystem', '-V'])

        try:
            output = subprocess.check_output(['./build/myexe/default/bin/myexe', 'build'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        self.assertEqual(output, '"build" is a directory')

    def test_c_library(self):
        clang_build_try_except(['-d', 'test/c-library', '-V'])

        try:
            output = subprocess.check_output(['./build/myexe/default/bin/myexe'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        self.assertEqual(output, '3 2 0'+os.linesep+'3 1 0')

    def test_build_all(self):
        clang_build_try_except(['-d', 'test/c-library', '-V', '-a'])

        try:
            output = subprocess.check_output(['./build/qhull/qhull-executable/default/bin/qhull', '-V'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail('Could not run a target which should have been built')

        self.assertEqual(output, 'qhull_r 7.2.0 (2015.2.r 2016/01/18)')

    def test_platform_flags(self):
        clang_build_try_except(['-d', 'test/platform_flags', '-V', '--debug'])

        try:
            output = subprocess.check_output(['./build/default/bin/myexe'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        if platform.PLATFORM == 'linux':
            self.assertEqual(output, 'Hello Linux!')
        elif platform.PLATFORM == 'osx':
            self.assertEqual(output, 'Hello OSX!')
        elif platform.PLATFORM == 'windows':
            self.assertEqual(output, 'Hello Windows!')
        else:
            from sys import platform as _platform
            raise RuntimeError('Tried to run test_platform_flags on unsupported platform ' + _platform)

    def test_openmp(self):
        clang_build_try_except(['-d', 'test/openmp', '-V'])

        try:
            output = subprocess.check_output(['./build/default/bin/runHello'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        self.assertRegex(output, r'Hello from thread 1, nthreads*')

    def test_mwe_two_targets(self):
        clang_build_try_except(['-d', 'test/multi_target_external', '-V', '--bundle'])

        try:
            output = subprocess.check_output(['./build/myexe/default/bin/runLib'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            self.fail(f'Could not run compiled program. Message:\n{e.output}')

        self.assertEqual(output, 'Hello! mylib::calculate() returned 2')

    def test_pybind11(self):
        clang_build_try_except(['-d', 'test/pybind11', '-V'])

        pylib_dir = os.path.abspath(os.path.join("build", "pylib", "default", platform.SHARED_LIBRARY_OUTPUT))
        sys.path.insert(0, pylib_dir)

        try:
            import pylib
            output = pylib.triple(3)
            self.assertEqual(output, 9)

        except ImportError:
            if os.path.exists(pylib_dir):
                print(f'Expected location "{pylib_dir}" contains: {os.listdir(pylib_dir)}')
            else:
                print(f'Expected location "{pylib_dir}" does not exist!')
            self.fail('Import of pylib failed!')

    def setUp(self):
        logger = logging.getLogger('clang_build')
        logger.setLevel(logging.INFO)
        ch = TqdmHandler()
        formatter = logging.Formatter('%(message)s')
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        logger.handlers = []
        logger.addHandler(ch)

    def tearDown(self):
        if _Path('build').exists():
            shutil.rmtree('build', onerror = on_rm_error)


if __name__ == '__main__':
    freeze_support()
    unittest.main()


"""Module containing compiler specifics."""

import logging as _logging
import shutil as _shutil
import subprocess as _subprocess
from functools import lru_cache as _lru_cache
from pathlib import Path as _Path
from re import search as _search

from .platform import PLATFORM_PYTHON_INCLUDE_PATH, PLATFORM_PYTHON_LIBRARY_PATH

_LOGGER = _logging.getLogger(__name__)


class Clang:
    """Class of the clang compiler and related tools.

    This class bundles various paths and attributes of
    the installed clang compiler. The clang version used
    is the one that would be used if `clang` or `clang++`
    were called in the terminal directly.

    Attributes
    ----------
    clang : :any:`pathlib.Path`
        Path to the `clang` executable
    clangpp : :any:`pathlib.Path`
        Path to the `clang++` executable
    clang_ar : :any:`pathlib.Path`
        Path to the `llvm-ar` executable
    max_cpp_dialect : str
        Compile flag for the latest supported
        dialect of the found compiler

    """

    _UNSUPPORTED_DIALECT_MESSAGE = "error: invalid value 'c++{0:02d}'"

    def __init__(self):
        """Search for clang and detect compiler features.

        Raises
        ------
        RuntimeError
            If a compiler or linker tool wasn't found on the system.

        """
        self.clang = self._find("clang")
        self.clangpp = self._find("clang++")
        self.clang_ar = self._find("llvm-ar")

        self.max_cpp_dialect = self._get_max_supported_compiler_dialect()

        _LOGGER.info("llvm root directory: %s", self.clangpp.parents[0])
        _LOGGER.info("clang executable:    %s", self.clang)
        _LOGGER.info("clang++ executable:  %s", self.clangpp)
        _LOGGER.info("llvm-ar executable:  %s", self.clang_ar)
        _LOGGER.info("Newest supported C++ dialect: %s", self.max_cpp_dialect)
        _LOGGER.info(f'Python headers in:   {PLATFORM_PYTHON_INCLUDE_PATH}')
        _LOGGER.info(f'Python library in:   {PLATFORM_PYTHON_LIBRARY_PATH}')

    def _find(self, executable):
        """Find path of executable.

        Parameters
        ----------
        executable : str
            The executable for which to search the location.

        Returns
        -------
        pathlib.Path
            Path where the executable was found.

        Raises
        ------
        RuntimeError
            If the executable was not found in the systems default
            look-up places.

        """
        try:
            return _Path(_shutil.which(executable))
        except TypeError:
            error_message = f"Couldn't find {executable} executable"
            _LOGGER.error(error_message)
            raise RuntimeError(error_message)

    def _get_dialect_flag(self, year):
        """Return a dialect flag for a given year.

        Parameters
        ----------
        year : int
            The last two digits of the year to convert.
            For example 11 will yield `-std=c++11`.

        Returns
        -------
        str
            A dialect flag from a given year.

        """
        return "-std=c++{:02d}".format(year)

    @_lru_cache(maxsize=1)
    def dialect_exists(self, year):
        """Check if a given dialect flag is valid.

        Parameters
        ----------
        year : int
            The last two digits of the dialect.
            For example 11 for `C++11`.

        Returns
        -------
        bool
            True if the dialect for the given year is supported
            by clang.

        """
        std_opt = self._get_dialect_flag(year)
        try:
            _subprocess.run(
                [str(self.clangpp), std_opt, "-x", "c++", "-E", "-"],
                check=True,
                input=b"",
                stdout=_subprocess.PIPE,
                stderr=_subprocess.PIPE,
            )
        except _subprocess.CalledProcessError as subprocess_error:
            error_message = self._UNSUPPORTED_DIALECT_MESSAGE.format(year)
            strerr = subprocess_error.stderr.decode("utf8", "strict")
            if error_message in strerr:
                return False
            raise
        return True

    @_lru_cache(maxsize=1)
    def _get_max_supported_compiler_dialect(self):
        """Check the maximally supported C++ dialect.

        Returns
        -------
        str
            Flag string of the latest supported dialect

        """
        try:
            _subprocess.run(
                [str(self.clangpp), "-std=dummpy", "-x", "c++", "-E", "-"],
                check=True,
                stdout=_subprocess.PIPE,
                stderr=_subprocess.PIPE,
                encoding="utf8",
            )
        except _subprocess.CalledProcessError as subprocess_error:
            for line in reversed(subprocess_error.stderr.splitlines()):
                if "draft" in line or "gnu" in line:
                    continue

                return "-std=" + _search(r"'(c\+\+..)'", line).group(1)

            raise RuntimeError("Could not find a supported C++ standard.")



"""
This module contains the `Environment` class.
"""

import logging as _logging
import shutil as _shutil
from multiprocessing import Pool as _Pool
from pathlib import Path as _Path

from .build_type import BuildType as _BuildType
from .clang_build import __version__
from .compiler import Clang as _Clang

_LOGGER = _logging.getLogger(__name__)


class Environment:
    """
    """

    def __init__(self, args):

        # TODO: Move this out
        _LOGGER.info(f"clang-build {__version__}")

        self.compiler = _Clang()

        # Build type (Default, Release, Debug)
        self.build_type = args.get("build_type", _BuildType.Default)
        _LOGGER.info(f"Build type: {self.build_type.name}")

        # Whether to force a rebuild
        self.force_build = args.get("force_build", False)
        if self.force_build:
            _LOGGER.info("Forcing rebuild...")

        # Build directory
        self.build_directory = _Path("build")

        # Whether to create a dotfile for graphing dependencies
        self.create_dependency_dotfile = not args.get("no_graph", False)

        # Whether to recursively clone submodules when cloning with git
        self.clone_recursive = not args.get("no_recursive_clone", False)

        # Whether to bundle binaries
        self.bundle = args.get("bundle", False)
        if self.bundle:
            _LOGGER.info("Bundling of binary dependencies is activated")

        # Whether to create redistributable bundles
        self.redistributable = args.get("redistributable", False)
        if self.redistributable:
            self.bundle = True
            _LOGGER.info("Redistributable bundling of binary dependencies is activated")



import clang_build

def get_project(directory, environment, parent=None) -> clang_build.project.Project:

    py_include_dir = clang_build.platform.PLATFORM_PYTHON_INCLUDE_PATH
    py_library_dir = clang_build.platform.PLATFORM_PYTHON_LIBRARY_PATH
    py_library_name = clang_build.platform.PLATFORM_PYTHON_LIBRARY_NAME
    py_library_extension = clang_build.platform.PLATFORM_PYTHON_EXTENSION_SUFFIX

    project = clang_build.project.Project("", {}, directory, environment, parent=parent)

    pylib = clang_build.target.TargetDescription(
        "pylib",
        {
            "target_type": "shared library",
            "output_name": "pylib",
            "output_prefix": "",
            "output_suffix": py_library_extension,
            "dependencies": ["pybind", "mylib"],
            "sources": ["python/bindings.cpp"],
            "include_directories": [py_include_dir],
            "flags": {
                "compile": ['-Wno-deprecated-declarations'],
                "link": ['-Wno-deprecated-declarations', f'-L {py_library_dir}', f'-l{py_library_name}']
            },
            "windows": {
                "flags": {
                    "compile": ['-Dstrdup=_strdup']
                }
            },
        },
        project
    )

    mylib = clang_build.target.TargetDescription(
        "mylib",
        {
            "target_type": "static library",
            "output_name": "mylib",
            "sources": ["src/mylib.cpp"]
        },
        project
    )

    pybind = clang_build.target.TargetDescription(
        "pybind",
        {
            "target_type": "header only",
            "url": "https://github.com/pybind/pybind11"
        },
        project
    )

    project.add_targets([mylib, pylib, pybind])

    return project


"""Module for the TreeEntry class."""


class TreeEntry:
    """Abstract class to make an object insertable into a tree.

    Implements means to identify and compare an object
    based on a member variable called ``identifier``
    which every subclass has to define.
    """

    identifier = None

    def __eq__(self, other) -> bool:
        """Declare two objects identical if ``identifier`` is the same."""
        if isinstance(other, str):
            return self.identifier == other

        return self.identifier == other.identifier

    def __hash__(self) -> int:
        """Return a hash value.

        Hashes the ``identifier`` value of this
        object making every object with the same
        ``identifier`` equal to this one (in terms
        of hash value).
        """
        return hash(self.identifier)


"""A class that contains potentially multiple targets and other projects."""

import logging as _logging
import textwrap as _textwrap
from multiprocessing import Pool as _Pool
from pathlib import Path as _Path
from typing import Optional as _Optional
from importlib import util as importlib_util

import networkx as _nx
import toml

from .circle import Circle as _Circle
from .io_tools import get_sources_and_headers as _get_sources_and_headers
from .logging_tools import NamedLogger as _NamedLogger
from .target import TARGET_MAP as _TARGET_MAP
from .target import Target as _Target
from .target import Executable as _Executable
from .target import HeaderOnly as _HeaderOnly
from .target import TargetDescription as _TargetDescription
from .tree_entry import TreeEntry as _TreeEntry
from .git_tools import download_sources as _git_download_sources
from .git_tools import get_latest_changes as _get_latest_changes

_LOGGER = _logging.getLogger(__name__)


def _get_project(directory, environment, parent=None):
    """Returns a Project created from a "clang-build.py" script.

    This is the basis of the scripting API. The "clang-build.py" script is required to
    define a method `get_project(directory, environment, parent) -> Project`.
    """
    module_file_path = directory.resolve() / "clang-build.py"
    module_name = "clang-build"

    module_spec = importlib_util.spec_from_file_location(module_name, module_file_path)
    if module_spec is None:
        raise RuntimeError(f'No "{module_name}" module could be found in "{directory.resolve()}"')

    clang_build_module = importlib_util.module_from_spec(module_spec)
    module_spec.loader.exec_module(clang_build_module)

    if clang_build_module.get_project is None:
        raise RuntimeError(f'Module "{module_name}" in "{directory.resolve()}" does not contain a `get_project` method')

    return clang_build_module.get_project(directory, environment, parent=parent)


class Project(_NamedLogger, _TreeEntry):
    """The Project class is all about organising targets and managing folder structures.

    Every start of clang-build, a Project is created that scans a given folder for what
    could potentially be a large multi-subproject structure. The project will find and
    configure all targets as specified by the parameters given. Once it is created, the
    structure that was discovered can be explored and targets can be built.
    """

    @property
    def build_directory(self):
        """Return the output directory for build objects for this project.

        Apart from a folder where the project files are stored, a project
        also has a separate folder where build files (like compiled object
        files or executables, etc.) are stored.
        """
        return self._build_directory

    @property
    def config(self):
        """Return the config dict of this project.

        Non-trivial projects will normally define project details in a toml file.
        This dict is a :any:`dict` version of this file.
        """
        return self._config

    @property
    def directory(self):
        """Return the directory of this project.

        Every project is located in a directory that has to be specified.
        The directory is the one that either contains a toml file or for
        trivial projects just the default folder structure that can be
        searched by clang-build.
        """
        return self._directory

    @property
    def environment(self):
        """Return the set of global settings."""
        return self._environment

    @property
    def identifier(self) -> str:
        """Returns the unique identifier of this project.

        The unique identifier for a project can be:
        - "project" if this is a basic project and no name was given
        - A simple string if a name was given to this project
        - A string with one or more "." if this is a subproject
        """
        return (
            self.name
            if self._parent is None
            else f"{self._parent.identifier}.{self.name}"
        )

    @property
    def name(self) -> str:
        """Return the name of this project.

        Every project can be given a name. A name is a string
        that does not contain "."s. For basic project no name
        has to be provided.
        """
        return self._name

    @property
    def parent(self):
        """Return the parent of this project.

        Projects are organised in a tree like structure. This project
        will return the parent project except for the case where this
        project is the top level project.
        """
        return self._parent

    @property
    def project_tree(self):
        """Return a :any:`networkx` tree representation of all selected targets.

        Targets can have dependencies on each other. This can be represented
        in a tree structure. Therefore, the targets that were selected during
        the initialization of the Project are available as a DiGraph. This
        graph is the global project tree. If you are only interested in subgraphs
        you have to use the :any:`networkx` functionality.
        """
        return self._project_tree

    @property
    def subprojects(self):
        """Return a list of subprojects.

        All direct children of this project (subprojects only, no targets)
        are returned as a list.
        """
        return self._subprojects

    @property
    def target_list(self):
        """Return all targets that are defined in this project.

        This list includes only those targets that were defined in this
        project and not those defined in subprojects.
        """
        return self._current_targets

    def __init__(self, name, config, directory, environment, **kwargs):
        """Initialise a project.

        The procedure for initialisation is:

        #. Setting some instance attributes
        #. Initialising sub-projects and filling the dependency tree recursively
        #. Determining the targets to configure
        #. Configuring the targets

        Parameters
        ----------
        directory : str or :any:`pathlib.Path`
            The directory to search for a ``toml`` file or source files
        environment : any:`clang_build.clang_build.Environment`
            An any:`clang_build.clang_build.Environment` instance defining some global
            settings for this run of clang-build.
        kwargs
            Used for internal purposes only.
        """
        _NamedLogger.__init__(self, _LOGGER)
        self._name = name
        self._config = config
        self._directory = _Path(directory)
        if not self._directory.exists():
            error_message = (
                f"ERROR: specified non-existent directory '{self._directory}'"
            )
            _LOGGER.error(error_message)
            raise RuntimeError(error_message)

        self._environment = environment
        self._parent = kwargs.get("parent", None)
        self._current_targets = []

        if not self.name:
            if self.parent:
                error_message = self.parent.log_message(
                    f"the config defines a subproject in folder {self.directory},"
                    + " that does not specify a name. It is not allowed to add a"
                    + " subproject that does not have a name."
                )

                self._logger.exception(error_message)
                raise RuntimeError(error_message)

            if "subprojects" in self.config:
                error_message = self.log_message(
                    "defining a top-level project with subprojects but without a name"
                    + " is illegal."
                )

                self._logger.exception(error_message)
                raise RuntimeError(error_message)

            self._name = "project"

        if self._parent:
            self._project_tree = self._parent.project_tree
        else:
            self._project_tree = _nx.DiGraph()

        self._set_directories()

        self._subprojects = self._parse_subprojects()

        self._project_tree.add_node(self, data=self)
        # Add edges to subprojects
        self._project_tree.add_edges_from(
            (self, sub_project) for sub_project in self.subprojects
        )

        # TODO: the following function name `_get_target_descriptions` is not descriptive
        self.add_targets(self._get_target_descriptions() + kwargs.get("targets", []))

    def __repr__(self) -> str:
        return f"clang_build.project.Project('{self.identifier}')"

    def __str__(self) -> str:
        return f"[[{self.identifier}]]"

    @classmethod
    def from_directory(cls, directory, environment, **kwargs):
        """Generate a project from a directory.

        This method covers the following cases in order:
        1. initialization from a "clang-build.py", if present
        2. initialization from a "clang-build.toml", if present
        3. defaults without any configuration
        """
        ### ----- Get the config
        toml_file = directory / "clang-build.toml"
        py_file = directory / "clang-build.py"
        parent = kwargs.get("parent", None)

        if parent:
            logger = parent._logger
            message_suffix = " for subproject"
        else:
            logger = _LOGGER
            message_suffix = ""

        if py_file.exists():
            logger.info(f"Using python project file \"{py_file}\"{message_suffix}")
            project = _get_project(directory, environment, parent=parent)
            if not isinstance(project, Project):
                raise RuntimeError(f'Unable to initialize project:\nThe `get_project` method in "{py_file}" did not return a valid `clang_build.project.Project`, its type is "{type(project)}"')
            return project

        elif toml_file.exists():
            logger.info(f"Found config file '{toml_file}'.")
            config = toml.load(toml_file)

        elif parent:
            error_message = parent.log_message(
                f"the config defines a subproject in folder {directory},"
                + " which does not have a project file. It is not allowed to add a"
                + " subproject that does not have a project file."
            )
            logger.exception(error_message)
            raise RuntimeError(error_message)

        else:
            config = {"target": {"output_name": "main"}}
        ### -----

        return cls.from_config(config, directory, environment, **kwargs)

    @classmethod
    def from_config(cls, config: dict, directory, environment, **kwargs):
        """Generate a project from a config dictionary.

        optional kwarg `targets`: Any of the `targets` may be a `Target` or `TargetDescription`
        """
        name = str(config.get("name", ""))
        project = cls(name, config, directory, environment, **kwargs)

        return project

    def add_targets(self, target_list: list):
        """Add a list of targets to this project.

        This method does the following:
        - check the validity of targets in the given list
        - add to the project's target_list
        - amend the dependency tree:
          - Add nodes and edges for targets in self
          - Add edges for dependencies in targets defined in project
        - optionally create a new dotfile
        - check for circular dependencies in the project

        This method integrates this project into the global project tree.

        This helper function is part of the initialisation of a project.
        It adds this project, all targets and their dependencies to the
        global project tree. If there are illegal dependencies, this function
        will raise an exception.
        """
        for target in target_list:
            if not (isinstance(target, _TargetDescription) or isinstance(target, _Target)):
                raise RuntimeError(f"clang_build.project.Project.add_targets: cannot add instance of type {type(target)}, it should be either a TargetDescription or Target.")

        self._current_targets += target_list

        # Add nodes and edges for targets in self
        for target in target_list:
            self._project_tree.add_node(target, data=target)
        self._project_tree.add_edges_from((self, target) for target in target_list)

        # Create a dotfile of the dependency graph
        create_dotfile = False
        if self._environment.create_dependency_dotfile and not self._parent:
            try:
                import pydot
                create_dotfile = True
            except:
                self._logger.error(f'Could not create dependency dotfile, as pydot is not installed')

        # Create initial dotfile without full dependency resolution
        if create_dotfile:
            _Path(self._environment.build_directory).mkdir(parents=True, exist_ok=True)
            _nx.drawing.nx_pydot.write_dot(self._project_tree, str(self._environment.build_directory / 'dependencies.dot'))

        # Add edges for dependencies in targets defined in project
        for target in target_list:
            dependencies = target.config.get("dependencies", [])
            dependency_objs = []
            for dependency_name in dependencies:
                full_name = self._identifier_from_name(dependency_name)

                try:
                    dependency = self._project_tree.nodes[full_name]["data"]
                except KeyError:
                    error_message = target.log_message(
                        f"the dependency [{dependency_name}] (expanded to [{full_name}]) does not point to a valid target."
                    )
                    self._logger.exception(error_message)
                    raise RuntimeError(error_message)

                dependency_objs.append(dependency)
                self._project_tree.add_edge(target, dependency)

            target.config["dependencies"] = dependency_objs

        # Create new dotfile with full dependency graph
        if create_dotfile:
            _nx.drawing.nx_pydot.write_dot(self._project_tree, str(self._environment.build_directory / 'dependencies.dot'))

        # Check the dependency graph for cycles
        if not self.parent:
            self._check_for_circular_dependencies()

        # Update the `only_target` flag of the targets in this project
        n_targets = len(self._current_targets)
        only_target = n_targets == 1 and len(self.config.get("subprojects", [])) == 0

        for target in self._current_targets:
            target.only_target = only_target


    def _parse_subprojects(self):
        """Generate all subprojects recursively.

        Recursively walks through all subprojects and parses
        them.

        Returns
        -------
        list
            List of Project objects.
        """
        subproject_list = []
        for directory in self.config.get("subprojects", []):
            subproject_dir = self._directory / directory
            subproject_list.append(Project.from_directory(subproject_dir, self._environment, parent=self))
        return subproject_list

    def _get_target_descriptions(self):
        """Get the list of targets for this project.

        Will return only those targets that are defined in
        this project. No subproject targets will be returned.

        Returns
        -------
        list
            List of TargetDescription objects
        """
        return [
            _TargetDescription(
                key, val, self
            )
            for key, val in self.config.items()
            if isinstance(val, dict)
        ]

    def _check_for_circular_dependencies(self):
        """Check if targets are defined in a circular dependency.

        Raises an exception if circular dependencies were found pointing
        out where the circular dependencies were found.
        """
        circles = list(_nx.simple_cycles(self._project_tree))

        circles = [_Circle(circle + [circle[0]]) for circle in circles]
        if circles:
            error_message = "Found the following circular dependencies:\n" + _textwrap.indent(
                    "\n".join("- " + str(circle) for circle in circles), prefix=" " * 3)
            self._logger.exception(error_message)
            raise RuntimeError(self.log_message(error_message))

    def build(
        self,
        build_all: bool = False,
        target_list: _Optional[list] = None,
        number_of_threads: _Optional[int] = None,
    ):
        """Build targets of this project.

        By default, this function builds all targets in this project as well
        as all their dependencies. This function will configure all targets
        that haven't been configured in a previous call.

        Parameters
        ----------
        build_all : bool
            If set to true, will not only build all targets in this project
            and their dependencies, but also all targets of all sub-projects.
        target_list : list
            If given, will build all targets in this project that are in the
            given list, as well as all their dependencies.
        number_of_threads : int
            If given will compile targets with the given number of threads. Otherwise
            it will use the default number of CPU cores visible to Python.

        """
        # Get targets to build
        targets_to_build = self._get_targets_to_build(build_all, target_list)

        # Sort targets in build order
        target_build_description_list = [
            target
            for target in reversed(list(_nx.topological_sort(self._project_tree)))
            if target in targets_to_build
            and not isinstance(self._project_tree.nodes[target]["data"], Project)
        ]

        # Get project sources, if any
        project_build_list = []
        for target_description in target_build_description_list:
            for predecessor in self._project_tree.predecessors(target_description):
                if isinstance(predecessor, Project):
                    project_build_list.append(predecessor)
        project_build_list = list(dict.fromkeys(project_build_list))
        for project in project_build_list:
            project.get_sources()

        for target in target_build_description_list:
            parent_project = next(self._project_tree.predecessors(target))
            target.parent_directory = parent_project._directory
            target.parent_build_directory = parent_project._build_directory

        ### Note: the project_tree needs to be updated directly for dependencies
        ### to be used correctly in the `_target_from_description` function
        target_build_list = []
        for target in target_build_description_list:
            if isinstance(target, _TargetDescription):
                target_instance = self._target_from_description(
                        self._project_tree.nodes[target]["data"]#, self._project_tree
                    )
                target_build_list.append(target_instance)
                self._project_tree.nodes[target]["data"] = target_instance
            else:
                target_build_list.append(target)

        if not target_build_list:
            self._logger.info("No targets to be built")

        # Compile
        with _Pool(processes=number_of_threads) as process_pool:
            for target in target_build_list:
                target.compile(process_pool, False)

        # Link
        for target in target_build_list:
            target.link()

        # Bundle
        if self._environment.bundle:
            with _Pool(processes=number_of_threads) as process_pool:
                for target in target_build_list:
                    target.bundle()

        # Redistributable bundle
        if self._environment.redistributable:
            with _Pool(processes=number_of_threads) as process_pool:
                for target in target_build_list:
                    target.redistributable()

    def _get_targets_to_build(
        self, build_all: bool = False, target_list: _Optional[list] = None
    ):
        """Return the list of targets to configure.

        This helper function is part of the initialisation of a project.
        Depending on the settings passed to clang build, not all targets
        will be configured. This function returns the list of only those
        targets that were selected and those targets that are dependencies
        of the selected ones.
        """
        if build_all:
            build_descendants_of = [self]
            _LOGGER.info("Building all targets...")
            if target_list:
                error_message = f"Cannot build [all] targets and only [{target_list}]!"
                _LOGGER.error(error_message)
                raise ValueError(error_message)
        elif target_list:
            build_descendants_of = target_list
        else:
            build_descendants_of = self._current_targets

        targets_to_build = set().union(
            set(build_descendants_of),
            *[
                set(
                    _nx.get_node_attributes(
                        self._project_tree.subgraph(
                            _nx.descendants(self._project_tree, target)
                        ),
                        "data",
                    ).values()
                )
                for target in build_descendants_of
            ]
        )

        return [
            target
            for target in targets_to_build
            if not isinstance(self._project_tree.nodes[target]["data"], Project)
        ]

    def _identifier_from_name(self, target_name: str) -> str:
        """Convert a name into an identifier.

        This works only if the name is defined for this project. If the
        target name is from a subproject, the method of the sub-project
        has to be called.

        Parameters
        ----------
        target_name : str
            The target name of the target for which to get the identifier.
            Has to be a target of this project.

        """
        prefix = f"{self.identifier}." if self.identifier else ""

        return f"{prefix}{target_name}"

    def _set_directories(self):
        """Set the build file output directory.

        This helper function is part of the initialisation of a project.
        It determines the build output directory which depends on if there
        are subprojects and more than one target. If that is the case, all
        output files will be in a hirarchical structure identical to the
        project tree structure. Else, the output for a trivial project will
        be directly in the build directory.
        """
        self._build_directory = self.environment.build_directory
        if self.parent:
            self._build_directory = self.parent.build_directory / self.name
        if self._config.get("url", None):
            self._directory = self.build_directory / "external_sources" / str(self._config.get("directory", ""))

    def _set_name(self, name=""):
        """Set the name.

        This helper function is part of the initialisation of a project.
        It will try to load the name from the config. If it cannot find it
        it will make sure that not having a name is allowed which is only
        the case if:

        - This project is not a sub-project and
        - This project does not have any sub-projects
        """
        self._name = name

        if not self.name:
            if self.parent:
                error_message = self.parent.log_message(
                    f"the config defines a subproject in folder {self.directory},"
                    + " that does not specify a name. It is not allowed to add a"
                    + " subproject that does not have a name."
                )

                self._logger.exception(error_message)
                raise RuntimeError(error_message)

            if "subprojects" in self.config:
                error_message = self.log_message(
                    "defining a top-level project with subprojects but without a name"
                    + " is illegal."
                )

                self._logger.exception(error_message)
                raise RuntimeError(error_message)

            self._name = "project"

    def _get_dependencies(self, target_description):
        dependencies = [
            self._project_tree.nodes()[dependency]["data"]
            for dependency in self._project_tree.successors(target_description)
        ]

        # Are there executables named as dependencies?
        executable_dependencies = [
            target for target in dependencies if isinstance(target, _Executable)
        ]
        if executable_dependencies:
            exelist = ", ".join([f"[{dep.name}]" for dep in executable_dependencies])
            error_message = target_description.log_message(
                f"The following targets are linking dependencies but were identified as executables:\n    {exelist}"
            )
            self._logger.error(error_message)
            raise RuntimeError(error_message)

        return dependencies

    def _target_from_description(self, target_description):
        """Return the appropriate target given the target description.

        A target type can either be speicified in the ``target_description.config``
        or a target is determined based on which files are found on the hard disk.
        If only header files are found, a header-only target is assumed. Else,
        an executable target will be generated.

        Targets need to be built from a bottom-up traversal of the project tree so
        that all the dependencies of targets are already generated.

        Parameters
        ----------
        target_description : :any:`TargetDescription`
            A shallow description class of the target
        project_tree : :any:`networkx.DiGraph`
            The entire project tree from which to extract
            dependencies

        Returns
        -------
        HeaderOnly or Executable or SharedLibrary or StaticLibrary
            Returns the correct target type based on input parameters
        """

        dependencies = self._get_dependencies(target_description)#, self._project_tree)

        # Sources
        target_description.get_sources()
        files = _get_sources_and_headers(
            target_description.name,
            target_description.config,
            target_description.root_directory,
            target_description.build_directory,
        )

        # Create specific target if the target type was specified
        target_type = target_description.config.get("target_type")
        if target_type is not None:
            target_type = str(target_type).lower()
            if target_type in _TARGET_MAP:
                return _TARGET_MAP[target_type](target_description, files, dependencies)
            else:
                error_message = target_description.log_message(
                    f'ERROR: Unsupported target type: "{target_description.config["target_type"].lower()}"'
                )
                self._logger.exception(error_message)
                raise RuntimeError(error_message)

        # No target specified so must be executable or header only
        else:
            if not files["sourcefiles"]:
                target_description.log_message(
                    "no source files found. Creating header-only target."
                )
                return _HeaderOnly(target_description, files, dependencies)

            target_description.log_message(
                f'{len(files["sourcefiles"])} source file(s) found. Creating executable target.'
            )
            return _Executable(target_description, files, dependencies)

    def get_sources(self):
        """External sources, if present, will be downloaded to build_directory/external_sources.
        """
        url = self._config.get("url", None)
        if url:
            version = self._config.get("version", None)
            download_directory = self.build_directory / "external_sources"
            _git_download_sources(url, download_directory, self._logger, version, self._environment.clone_recursive)

            self._directory = download_directory / self._config.get("directory", "")


from .build_type import BuildType
from . import platform as _platform

class BuildFlags:
    DEFAULT_COMPILE_FLAGS = {
        BuildType.Default: ["-Wall", "-Wextra", "-Wpedantic", "-Wshadow", "-Werror"],
        BuildType.Release: ["-O3", "-DNDEBUG"],
        BuildType.RelWithDebInfo: ["-O3", "-g3", "-DNDEBUG"],
        BuildType.Debug: [
            "-Og",
            "-g3",
            "-DDEBUG",
            "-fno-optimize-sibling-calls",
            "-fno-omit-frame-pointer",
            "-fsanitize=address",
            "-fsanitize=undefined",
        ],
        BuildType.Coverage: [
            "-Og",
            "-g3",
            "-DDEBUG",
            "-fno-optimize-sibling-calls",
            "-fno-omit-frame-pointer",
            "-fsanitize=address",
            "-fsanitize=undefined",
            "--coverage",
            "-fno-inline",
        ],
    }
    DEFAULT_LINK_FLAGS = {
        BuildType.Debug: ["-fsanitize=address", "-fsanitize=undefined"],
        BuildType.Coverage: [
            "-fsanitize=address",
            "-fsanitize=undefined",
            "--coverage",
            "-fno-inline",
        ],
    }

    def __init__(
        self, build_type, default_compile_flags=False, default_link_flags=False
    ):
        # TODO: Store default flags separately
        self.compile_default = []
        self.link_default = []

        self.compile_private = []
        self.link_private = []

        self.compile_interface = []
        self.link_interface = []

        self.compile_public = []
        self.link_public = []

        self._build_type = build_type

        if default_compile_flags:
            self.compile_default = self.DEFAULT_COMPILE_FLAGS[BuildType.Default]
            # self.compile_private += DEFAULT_COMPILE_FLAGS.get(BuildType.Default, [])
            if build_type != BuildType.Default:
                self.compile_default = self.DEFAULT_COMPILE_FLAGS.get(
                    build_type, []
                )
                # self.compile_private += DEFAULT_COMPILE_FLAGS.get(build_type, [])

        if default_link_flags:
            self.link_default = self.DEFAULT_LINK_FLAGS.get(build_type, [])
            # self.link_private += DEFAULT_LINK_FLAGS.get(build_type, [])

    def make_private_flags_public(self):
        self.compile_public += self.compile_private
        self.compile_private = []
        self.link_public += self.link_private
        self.link_private = []

    def apply_public_flags(self, target):
        self.compile_private += target.build_flags.compile_public
        self.link_private += target.build_flags.link_public

    def forward_public_flags(self, target):
        self.compile_public += target.build_flags.compile_public
        self.link_public += target.build_flags.link_public

    def apply_interface_flags(self, target):
        self.compile_private += target.build_flags.compile_interface
        self.link_private += target.build_flags.link_interface

    def forward_interface_flags(self, target):
        self.compile_interface += target.build_flags.compile_interface
        self.link_interface += target.build_flags.link_interface

    def add_target_flags(self, config):
        # Own private flags
        cf, lf = self._parse_flags_config(config, "flags")
        self.compile_private += cf
        self.link_private += lf

        # Own interface flags
        cf, lf = self._parse_flags_config(config, "interface-flags")
        self.compile_interface += cf
        self.link_interface += lf

        # Own public flags
        cf, lf = self._parse_flags_config(config, "public-flags")
        self.compile_public += cf
        self.link_public += lf

    def add_bundling_flags(self):
        self.link_private += _platform.PLATFORM_BUNDLING_LINKER_FLAGS

    def final_compile_flags_list(self):
        # TODO: Add max_dialect and plattform specific flags here as well
        #       Need to see how we get around the target-type-specific flags issue
        return list(dict.fromkeys(self.compile_private + self.compile_public))

    def final_link_flags_list(self):
        return list(dict.fromkeys(self.link_private + self.link_public))

    def _parse_flags_config(self, options, flags_kind='flags'):
        flags_dicts   = []
        compile_flags = []
        link_flags    = []

        if flags_kind in options:
            flags_dicts.append(options.get(flags_kind, {}))

        flags_dicts.append(options.get(_platform.PLATFORM, {}).get(flags_kind, {}))

        for fdict in flags_dicts:
            compile_flags += fdict.get('compile', [])
            link_flags    += fdict.get('link', [])

            if self._build_type != BuildType.Default:
                compile_flags += fdict.get(f'compile_{self._build_type}', [])

        return compile_flags, link_flags



import logging as _logging
import subprocess as _subprocess


def needs_download(url, download_directory, logger, version=None):
    if download_directory.exists():
        if download_directory.is_dir():
            if any(download_directory.iterdir()):
                try:
                    local_url = _subprocess.check_output(
                        ["git", "config", "--get", "remote.origin.url"],
                        stderr=_subprocess.PIPE,
                        encoding="utf-8",
                        cwd=download_directory,
                    ).splitlines()[0]
                except _subprocess.CalledProcessError:
                    return True

                if local_url != url:
                    logger.debug(
                        f"External sources folder exists but local url='{local_url}', while target url='{url}'"
                    )
                    return True

                if version:
                    checkout_version(version, download_directory, url, logger)
                else:
                    get_latest_changes(download_directory, logger)

                logger.debug(
                    f"External sources found in '{download_directory.resolve()}'"
                )

                return False
        else:
            error_message = f"Tried to download '{url}' to '{download_directory.resolve()}', but '{download_directory.resolve()}' is a file."
            raise RuntimeError(error_message)

    return True


def clone_repository(url, download_directory, recursive, logger):
    logger.debug(f"Downloading external sources to '{download_directory}'")
    download_directory.mkdir(parents=True, exist_ok=True)
    try:
        clone_command = ["git", "clone"]

        if recursive:
            clone_command += ["--recurse-submodules"]

        _subprocess.run(
            clone_command + [url, str(download_directory)],
            stdout=_subprocess.PIPE,
            stderr=_subprocess.PIPE,
            encoding="utf-8",
        )
    except _subprocess.CalledProcessError as e:
        error_message = (
            f"Error trying to download external target. Message " + e.output
        )
        logger.exception(error_message)
        raise RuntimeError(error_message)

    logger.debug(f"External sources downloaded")


def checkout_version(version, repository, url, logger):
    try:
        _subprocess.run(
            ["git", "fetch"],
            cwd=repository,
            stdout=_subprocess.PIPE,
            stderr=_subprocess.PIPE,
            encoding="utf-8",
        )
        _subprocess.run(
            ["git", "checkout", version],
            cwd=repository,
            stdout=_subprocess.PIPE,
            stderr=_subprocess.PIPE,
            encoding="utf-8",
        )
    except _subprocess.CalledProcessError as e:
        error_message = (
            f"Error trying to checkout version '{version}' from url '{url}'. Message "
            + e.output
        )
        logger.exception(error_message)
        raise RuntimeError(error_message)


def get_latest_changes(repository, logger):
    try:
        _subprocess.run(
            ["git", "pull"],
            cwd=repository,
            stdout=_subprocess.PIPE,
            stderr=_subprocess.PIPE,
            encoding="utf-8",
        )
    except _subprocess.CalledProcessError as e:
        error_message = (
            f"Unabled to get latest changes in repository '{repository}'. Message "
            + e.output
        )
        logger.exception(error_message)
        raise RuntimeError(error_message)


def download_sources(url, directory, logger, version=None, clone_recursively=True):
    """Download sources using git.
    """
    # Check if directory is already present and non-empty
    if needs_download(url, directory, logger, version):
        logger.info(
            f"downloading external sources to '{str(directory.resolve())}'"
        )
        clone_repository(
            url, directory, clone_recursively, logger
        )
        if version:
            checkout_version(version, directory, url, logger)
        else:
            get_latest_changes(directory, logger)

    # Otherwise we download the sources
    else:
        logger.debug(f"external sources found in '{str(directory.resolve())}'")


'''
clang-build:
  TODO: module docstring...
'''

import sys as _sys
import argparse as _argparse
import logging as _logging
from pathlib import Path as _Path
from multiprocessing import freeze_support as _freeze_support

from pbr.version import VersionInfo as _VersionInfo
_v = _VersionInfo('clang-build').semantic_version()
__version__ = _v.release_string()
version_info = _v.version_tuple

from .build_type import BuildType as _BuildType
from .project import Project as _Project
from .progress_bar import CategoryProgress as _CategoryProgress
from .logging_tools import TqdmHandler as _TqdmHandler
from .environment import Environment as _Environment
from .errors import CompileError as _CompileError
from .errors import LinkError as _LinkError
from .errors import BundleError as _BundleError
from .errors import RedistributableError as _RedistributableError

_LOGGER = _logging.getLogger(__name__)

def _setup_logger(log_level=None):
    _LOGGER.setLevel(_logging.DEBUG)

    # create formatter _and add it to the handlers
    formatter = _logging.Formatter('%(message)s')

    # add log file handler
    fh = _logging.FileHandler('clang-build.log', mode='w')
    fh.setLevel(_logging.DEBUG)
    fh.setFormatter(formatter)
    _LOGGER.addHandler(fh)

    if log_level is not None:
        ch = _TqdmHandler()
        ch.setLevel(log_level)
        ch.setFormatter(formatter)
        _LOGGER.addHandler(ch)


def parse_args(args):
    _command_line_description = (
    '`clang-build` is a build system to build your C++ projects. It uses the clang '
    'compiler/toolchain in the background and python as the build-system\'s scripting '
    'language.\n'
    'For more information please visit: https://github.com/trick-17/clang-build')
    parser = _argparse.ArgumentParser(description=_command_line_description, formatter_class=_argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-V', '--verbose',
                        help='activate more detailed output',
                        action='store_true')
    parser.add_argument('-p', '--progress',
                        help='activates a progress bar output',
                        action='store_true')
    parser.add_argument('-d', '--directory',
                        type=_Path,
                        help='set the root source directory')
    parser.add_argument('-b', '--build-type',
                        choices=list(_BuildType),
                        type=_BuildType,
                        default=_BuildType.Default,
                        help='set the build type for this project')
    target_group = parser.add_mutually_exclusive_group()
    target_group.add_argument('-a', '--all',
                        help='build every target, irrespective of whether any root target depends on it',
                        action='store_true')
    target_group.add_argument('-t', '--targets',
                        type=str,
                        nargs="+",
                        help='only these targets and their dependencies should be built')
    parser.add_argument('-f', '--force-build',
                        help='also build sources which have already been built',
                        action='store_true')
    parser.add_argument('-j', '--jobs',
                        type=int,
                        default=1,
                        help='set the number of concurrent build jobs')
    parser.add_argument('--debug',
                        help='activates additional debug output, overrides verbosity option.',
                        action='store_true')
    parser.add_argument('--no-graph',
                        help='deactivates output of a dependency graph dotfile',
                        action='store_true')
    parser.add_argument('--no-recursive-clone',
                        help='deactivates recursive cloning of git submodules',
                        action='store_true')
    parser.add_argument('--bundle',
                        help='automatically gather dependencies into the binary directories of targets',
                        action='store_true')
    parser.add_argument('--redistributable',
                        help='Automatically create redistributable bundles from binary bundles. Implies `--bundle`',
                        action='store_true')
    return parser.parse_args(args=args)


def build(args):
    # Create container of environment variables
    environment = _Environment(vars(args))

    categories = ['Configure', 'Compile', 'Link']
    if environment.bundle:
        categories.append('Generate bundle')
    if environment.redistributable:
        categories.append('Generate redistributable')

    with _CategoryProgress(categories, not args.progress) as progress_bar:
        project = _Project.from_directory(args.directory, environment)
        project.build(args.all, args.targets, args.jobs)
        progress_bar.update()

    _LOGGER.info('clang-build finished.')


def _main():
    # Build
    try:
        args = parse_args(_sys.argv[1:])

        # Logger verbosity
        if not args.debug:
            if args.verbose:
                _setup_logger(_logging.INFO)
            else:
                # Only file log
                _setup_logger(None)
        else:
            _setup_logger(_logging.DEBUG)

        build(args)

    except _CompileError as compile_error:
        _LOGGER.error('Compilation was unsuccessful:')
        for target, errors in compile_error.error_dict.items():
            printout = f'[{target}]: target did not compile. Errors:\n'
            printout += ' '.join(errors)
            _LOGGER.error(printout)
    except _LinkError as link_error:
        _LOGGER.error('Linking was unsuccessful:')
        for target, errors in link_error.error_dict.items():
            printout = f'[{target}]: target did not link. Errors:\n{errors}'
            _LOGGER.error(printout)
    except _BundleError as bundle_error:
        _LOGGER.error('Bundling was unsuccessful:')
        for target, errors in bundle_error.error_dict.items():
            printout = f'[{target}]: target could not be bundled. Errors:\n{errors}'
            _LOGGER.error(printout)
    except _RedistributableError as redistributable_error:
        _LOGGER.error('Redistibutable bundling was unsuccessful:')
        for target, errors in redistributable_error.error_dict.items():
            printout = f'[{target}]: target could not be bundled into a redistributable. Errors:\n{errors}'
            _LOGGER.error(printout)


if __name__ == '__main__':
    _freeze_support()
    _main()



import clang_build

"""
Sources can be generated and subsequentially a project created
"""

def get_project(directory, environment, parent=None) -> clang_build.project.Project:
    with open(directory / "version.h", 'w') as f:
        f.write(
            "#define VERSION_MAJOR 1\n"
            "#define VERSION_MINOR 2\n"
            "#define VERSION_PATCH 0")
    with open(directory / "main.cpp", "w") as fh:
        fh.write(
            "#include <iostream>\n"
            "#include \"version.h\"\n"
            "\n"
            "int main()\n"
            "{\n"
            "    std::cerr << \"the version is \" << VERSION_MAJOR << \".\" << VERSION_MINOR << \".\" << VERSION_PATCH << std::endl;\n"
            "    return 0;\n"
            "}")

    project = clang_build.project.Project("", {}, directory, environment, parent=parent)
    target = clang_build.target.TargetDescription("", {}, project)
    project.add_targets([target])

    return project


"""
Target describes a single build or dependency target with all needed paths and
a list of buildables that comprise it's compile and link steps.
"""

import logging as _logging
import shutil as _shutil
import subprocess as _subprocess
from abc import abstractmethod
from multiprocessing import freeze_support as _freeze_support
from pathlib import Path as _Path

from . import platform as _platform
from .directories import Directories
from .errors import BundleError as _BundleError
from .errors import CompileError as _CompileError
from .errors import LinkError as _LinkError
from .errors import RedistributableError as _RedistributableError
from .flags import BuildFlags
from .git_tools import download_sources as _git_download_sources
from .logging_tools import NamedLogger as _NamedLogger
from .progress_bar import get_build_progress_bar as _get_build_progress_bar
from .single_source import SingleSource as _SingleSource
from .tree_entry import TreeEntry as _TreeEntry

_LOGGER = _logging.getLogger(__name__)

class Target(_TreeEntry, _NamedLogger):
    """Base class for all kinds of target, whose sources have been gathered.

    Target instances are used in the `build` method of a project.
    """

    @property
    def name(self):
        """Return the name of this target.
        """
        return self._name

    @property
    def identifier(self):
        """Return the unique identifier of this target.

        Targets are identified by their parent projects and their name as
        "[project.subproject.target]".
        """
        return self._identifier

    @property
    def dependencies(self):
        """Return a list of any:`clang_build.target.Target`, which this
        target depends on.
        """
        return self._dependencies

    @property
    def root_directory(self):
        """Folders "include", "src", etc. are searched
        relative to this folder.
        """

        return self._root_directory

    @property
    def build_directory(self):
        """Return the directory which serves as the root build folder
        for this target.
        """
        return self._build_directory

    @property
    def headers(self):
        """Headers found for this project.
        """
        return self._headers

    @property
    def directories(self):
        """Return the any:`clang_build.directories.Directories` in use
        by this target.
        """
        return self._directories

    def __init__(self, target_description, files, dependencies=None):
        """Initialise a target.

        The procedure for initialisation is:

        #. Setting some instance attributes
        #. Initialising sub-projects and filling the dependency tree recursively
        #. Determining the targets to configure
        #. Configuring the targets

        Parameters
        ----------
        target_description : :any:`clang_build.target.TargetDescription`
            All the information on how to gather sources and build the target.
        environment : any:`clang_build.clang_build.Environment`
            An any:`clang_build.clang_build.Environment` instance defining some global
            settings for this run of clang-build.
        dependencies
            Optional. A list of any:`clang_build.target.Target` which this target
            depends on.
        """
        _NamedLogger.__init__(self, _LOGGER)
        self._name = target_description.name
        self._identifier = target_description.identifier
        self._environment = target_description.environment

        if dependencies is None:
            dependencies = []

        self._dependencies = dependencies

        # TODO: parse user-specified target version

        self._root_directory = _Path(target_description.root_directory)
        self._build_directory = target_description.build_directory

        self._headers = list(dict.fromkeys(files["headers"]))

        self._directories = Directories(files, self._dependencies)

        # Compile and link flags
        self._build_flags = self._get_default_flags()

        # Dependencies' flags
        for target in self.dependencies:
            # Header only libraries will forward all non-private flags
            self._add_dependency_flags(target)

        self._build_flags.add_target_flags(target_description.config)

    def __repr__(self) -> str:
        return f"clang_build.target.Target('{self.identifier}')"

    def __str__(self) -> str:
        return f"[{self.identifier}]"

    @abstractmethod
    def _get_default_flags(self):
        """Overload to return any:`clang_build.flags.BuildFlags`, which the target should use.
        """
        pass

    @abstractmethod
    def _add_dependency_flags(self, target):
        """Overload to add flags from dependencies of this target to its own.
        """
        pass

    @abstractmethod
    def compile(self, process_pool, progress_disabled):
        """Compile the target, if applicable.

        This produces an OS-dependent output in the build/bin folder.
        """
        pass

    @abstractmethod
    def link(self):
        """Link the target, if applicable.

        This produces an OS-dependent output in the corresponding build folder:
        - "bin" for executables and shared objects
        - "lib" for static libraries
        """
        pass

    def bundle(self):
        """For executable and shared library targets, bundle shared library
        dependencies into the binary output folder and amend the rpath if
        necessary.

        They can therefore be used without amending the system PATH or similar.
        """
        self.unsuccessful_bundle = False
        bundle_files = []
        for dependency in self.dependencies:
            bundle_files += dependency.bundle()
        return bundle_files

    def redistributable(self):
        """Create a redistributable bundle, suitable for installation.

        The redistributable bundle contains
        - an "include" folder with the public headers (preserving folder structure).
          Note that this includes the headers of public dependencies.
        - "bin" and "lib" folders containing compiled output of the target
          and its dependencies.
        """
        self.unsuccessful_redistributable = False

    @property
    def build_flags(self):
        """Return the any:`clang_build.flags.BuildFlags` of this target.
        """
        return self._build_flags


class HeaderOnly(Target):
    """HeaderOnly targets are the default target type when no source files are found.

    Header-only targets cannot have private compile flags, link flags or dependencies.
    They are automatically promoted to public instead.

    TODO: need to check whether "public" makes sense for header-only, when we have implemented "private" dependencies
    """

    def __init__(self, target_description, files, dependencies=None):
        """Initialise a header-only target.

        Header-only targets' private flags and include-directories are public.
        """
        super().__init__(
            target_description=target_description,
            files=files,
            dependencies=dependencies,
        )

        if files["sourcefiles"]:
            self._logger.info(
                f'{len(files["sourcefiles"])} source file(s) found for header-only target. '
                + "You may want to check your build configuration."
            )

        self._build_flags.make_private_flags_public()
        self._directories.make_private_directories_public()

    def link(self):
        self._logger.info("header-only target does not require linking.")

    def compile(self, process_pool, progress_disabled):
        self._logger.info("header-only target does not require compiling.")

    def _get_default_flags(self):
        """Return the default any:`clang_build.flags.BuildFlags` without compile or link flags.
        """
        return BuildFlags(self._environment.build_type)

    def _add_dependency_flags(self, target):
        """Forward dependencies' public and interface flags.
        """
        self._build_flags.forward_public_flags(target)
        self._build_flags.forward_interface_flags(target)


def generate_depfile_single_source(buildable):
    buildable.generate_depfile()
    return buildable


def compile_single_source(buildable):
    buildable.compile()
    return buildable


class Compilable(Target):
    """A compilable target will generate object files.
    """

    def __init__(
        self,
        target_description,
        files,
        link_command,
        output_folder,
        platform_flags,
        prefix,
        suffix,
        dependencies=None,
    ):
        super().__init__(
            target_description=target_description,
            files=files,
            dependencies=dependencies,
        )

        self.source_files = files["sourcefiles"]

        if not self.source_files:
            error_message = f"[{self.identifier}]: ERROR: Target was defined as a {self.__class__.__name__} but no source files were found"
            self._logger.error(error_message)
            raise RuntimeError(error_message)

        self.object_directory = (self.build_directory / "obj").resolve()
        self.depfile_directory = (self.build_directory / "dep").resolve()
        self.output_folder = (self.build_directory / output_folder).resolve()
        self.redistributable_folder = (self.build_directory / "redistributable").resolve()

        prefix = target_description.config.get("output_prefix", prefix)
        suffix = target_description.config.get("output_suffix", suffix)

        self.outname = target_description.config.get("output_name", self.name)
        self.outfilename = prefix + self.outname + suffix
        self.outfile = (self.output_folder / self.outfilename).resolve()

        compile_flags = self._build_flags.final_compile_flags_list()

        # Buildables which this Target contains
        self.include_directories_command = self._directories.include_command()

        self.buildables = [
            _SingleSource(
                environment=self._environment,
                source_file=source_file,
                platform_flags=platform_flags,
                current_target_root_path=self.root_directory,
                depfile_directory=self.depfile_directory,
                object_directory=self.object_directory,
                include_strings=self.include_directories_command,
                compile_flags=compile_flags,
            )
            for source_file in self.source_files
        ]

        # If compilation of buildables fail, they will be stored here later
        self._unsuccessful_compilations = []

        # Linking setup
        self.link_command = link_command + [str(self.outfile)]

    def _get_default_flags(self):
        """Return the default any:`clang_build.flags.BuildFlags` with compile flags but without link flags.
        """
        return BuildFlags(self._environment.build_type, default_compile_flags=True)

    def compile(self, process_pool, progress_disabled):
        """From the list of source files, compile those which changed or whose dependencies (included headers, ...) changed.
        """

        # Object file only needs to be (re-)compiled if the source file or headers it depends on changed
        if not self._environment.force_build:
            self.needed_buildables = [
                buildable for buildable in self.buildables if buildable.needs_rebuild
            ]
        else:
            self.needed_buildables = self.buildables

        # If the target was not modified, it may not need to compile
        if not self.needed_buildables:
            self._logger.info("target is already compiled")
            return

        self._logger.info(
            "target needs to build sources %s", [b.name for b in self.needed_buildables]
        )

        # Execute depfile generation command
        #
        #
        #    TODO: Remove hardcoded progress bar and use callback function instead
        #
        #
        self._logger.info(f"generate dependency files")
        for b in self.needed_buildables:
            self._logger.debug(" ".join(b.dependency_command))
        self.needed_buildables = list(
            _get_build_progress_bar(
                process_pool.imap(
                    generate_depfile_single_source, self.needed_buildables
                ),
                progress_disabled,
                total=len(self.needed_buildables),
                name=self.name,
            )
        )

        # Execute compile command
        self._logger.info("compile object files")
        for b in self.needed_buildables:
            self._logger.debug(" ".join(b.compile_command))
        self.needed_buildables = list(
            _get_build_progress_bar(
                process_pool.imap(compile_single_source, self.needed_buildables),
                progress_disabled,
                total=len(self.needed_buildables),
                name=self.name,
            )
        )

        # Catch compilation errors
        self._unsuccessful_compilations = [
            buildable
            for buildable in self.needed_buildables
            if (buildable.compilation_failed or buildable.depfile_failed)
        ]
        if self._unsuccessful_compilations:
            raise _CompileError('Compilation was unsuccessful',
                {self.identifier: [source.compile_report for source in self._unsuccessful_compilations]})

    def link(self):
        link_command = str(" ".join(dict.fromkeys(self.link_command)))
        self._logger.info(f'link -> "{self.outfile}"')
        self._logger.debug("    " + link_command)
        link_command = list(link_command.split())

        # Execute link command
        try:
            self.output_folder.mkdir(parents=True, exist_ok=True)
            self.link_report = (
                _subprocess.check_output(link_command, stderr=_subprocess.STDOUT)
                .decode("utf-8")
                .strip()
            )
            self.unsuccessful_link = False
        except _subprocess.CalledProcessError as error:
            self.unsuccessful_link = True
            self.link_report = error.output.decode("utf-8").strip()

        # Catch link errors
        if self.unsuccessful_link:
            raise _LinkError('Linking was unsuccessful',
                {self.identifier: self.link_report})


class Executable(Compilable):
    """Executable targets are the default target type when source files are found.

    An executable cannot be the dependency of another target.
    """

    def __init__(self, target_description, files, dependencies=None):
        """Initialise an executable target.
        """

        super().__init__(
            target_description=target_description,
            files=files,
            link_command=[str(target_description.environment.compiler.clangpp), "-o"],
            output_folder=_platform.EXECUTABLE_OUTPUT,
            platform_flags=_platform.PLATFORM_EXTRA_FLAGS_EXECUTABLE,
            prefix=_platform.EXECUTABLE_PREFIX,
            suffix=_platform.EXECUTABLE_SUFFIX,
            dependencies=dependencies,
        )

        ### Link self
        self.link_command += [
            str(buildable.object_file) for buildable in self.buildables
        ]

        ### Library dependency search paths
        for target in self.dependencies:
            if target.__class__ is not HeaderOnly:
                self.link_command.append("-L " + str(target.output_folder.resolve()))

        ### Bundling requires extra flags
        if self._environment.bundle:
            self._build_flags.add_bundling_flags()

        self.link_command += self._build_flags.final_link_flags_list()

        ### Link dependencies
        for target in self.dependencies:
            if target.__class__ is not HeaderOnly:
                self.link_command.append("-l" + target.outname)

    def bundle(self):
        self.unsuccessful_bundle = False

        ### Gather
        bundle_files = []
        for dependency in self.dependencies:
            bundle_files += dependency.bundle()

        ### Copy
        for bundle_file in bundle_files:
            try:
                _shutil.copy(bundle_file, self.output_folder)
            except _subprocess.CalledProcessError as error:
                self.unsuccessful_bundle = True
                self.bundle_report = error.output.decode("utf-8").strip()

        # Catch bundling errors
        if self.unsuccessful_bundle:
            raise _BundleError('Bundling was unsuccessful',
                {self.identifier: self.bundle_report})

        return [self.outfile] + bundle_files

    def redistributable(self):
        self.unsuccessful_redistributable = False
        if _platform.PLATFORM == "osx":
            appfolder = self.redistributable_folder / f"{self.outname}.app"
            binfolder = appfolder / "Contents"/ "MacOS"
            try:
                binfolder.mkdir(parents=True, exist_ok=True)
                with (appfolder / "Contents"/ "Info.plist").open(
                    mode="w"
                ) as plist:
                    plist.write(
                        f"""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
  <key>CFBundleGetInfoString</key>
  <string>{self.outname}</string>
  <key>CFBundleExecutable</key>
  <string>{self.outname}</string>
  <key>CFBundleIdentifier</key>
  <string>com.your-company-name.www</string>
  <key>CFBundleName</key>
  <string>{self.outname}</string>
  <key>CFBundleShortVersionString</key>
  <string>0.0</string>
  <key>CFBundleInfoDictionaryVersion</key>
  <string>6.0</string>
  <key>CFBundlePackageType</key>
  <string>APPL</string>
  <key>IFMajorVersion</key>
  <integer>0</integer>
  <key>IFMinorVersion</key>
  <integer>0</integer>
</dict>
</plist>"""
                    )
                _shutil.copy(self.outfile, binfolder)
                bundle_files = self.bundle()
                for bundle_file in bundle_files:
                    _shutil.copy(bundle_file, binfolder)
            except _subprocess.CalledProcessError as error:
                self.unsuccessful_redistributable = True
                self.redistributable_report = error.output.decode("utf-8").strip()
        elif _platform.PLATFORM == "linux":
            try:
                self.redistributable_folder.mkdir(parents=True, exist_ok=True)
                # TODO: gather includes and shared libraries
            except _subprocess.CalledProcessError as error:
                self.unsuccessful_redistributable = True
                self.redistributable_report = error.output.decode("utf-8").strip()
        elif _platform.PLATFORM == "windows":
            try:
                self.redistributable_folder.mkdir(parents=True, exist_ok=True)
                # TODO: gather includes and shared libraries
            except _subprocess.CalledProcessError as error:
                self.unsuccessful_redistributable = True
                self.redistributable_report = error.output.decode("utf-8").strip()

        # Check for redistibutable bundling errors
        if self.unsuccessful_redistributable:
            raise _RedistributableError('Creating redistributables was unsuccessful',
                {self.identifier: self.redistributable_report})

    def _get_default_flags(self):
        """Return the default any:`clang_build.flags.BuildFlags` with compile flags and link flags.
        """
        return BuildFlags(self._environment.build_type, default_compile_flags=True, default_link_flags=True)

    def _add_dependency_flags(self, target):
        """Add dependencies' public and interface flags to the own and forward their public flags.
        """
        self._build_flags.apply_public_flags(target)
        self._build_flags.forward_public_flags(target)
        self._build_flags.apply_interface_flags(target)


class SharedLibrary(Compilable):
    def __init__(self, target_description, files, dependencies=None):

        super().__init__(
            target_description=target_description,
            files=files,
            link_command=[str(target_description.environment.compiler.clangpp), "-shared", "-o"],
            output_folder=_platform.SHARED_LIBRARY_OUTPUT,
            platform_flags=_platform.PLATFORM_EXTRA_FLAGS_SHARED,
            prefix=_platform.SHARED_LIBRARY_PREFIX,
            suffix=_platform.SHARED_LIBRARY_SUFFIX,
            dependencies=dependencies,
        )

        ### Link self
        self.link_command += [
            str(buildable.object_file) for buildable in self.buildables
        ]

        ### Library dependency search paths
        for target in self.dependencies:
            if target.__class__ is not HeaderOnly:
                self.link_command.append("-L " + str(target.output_folder.resolve()))

        self.link_command += self._build_flags.final_link_flags_list()

        ### Link dependencies
        for target in self.dependencies:
            if target.__class__ is not HeaderOnly:
                self.link_command.append("-l" + target.outname)

        ### Bundling requires some link flags
        if self._environment.bundle:
            if _platform.PLATFORM == "osx":
                ### Install name for OSX
                self.link_command += [f"-install_name @rpath/{self.outfilename}"]
            elif _platform.PLATFORM == "linux":
                pass
            elif _platform.PLATFORM == "windows":
                pass

    def bundle(self):
        self.unsuccessful_bundle = False

        ### Gather
        self_bundle_files = [self.outfile]
        if _platform.PLATFORM == "windows":
            self_bundle_files.append(_Path(str(self.outfile)[:-3] + "exp"))
            self_bundle_files.append(_Path(str(self.outfile)[:-3] + "lib"))

        bundle_files = []
        for dependency in self.dependencies:
            bundle_files += dependency.bundle()

        ### Copy
        for bundle_file in bundle_files:
            try:
                _shutil.copy(bundle_file, self.output_folder)
            except _subprocess.CalledProcessError as error:
                self.unsuccessful_bundle = True
                self.bundle_report = error.output.decode("utf-8").strip()

        # Catch bundling errors
        if self.unsuccessful_bundle:
            raise _BundleError('Bundling was unsuccessful',
                {self.identifier: self.bundle_report})

        return self_bundle_files + bundle_files

    def _get_default_flags(self):
        """Return the default any:`clang_build.flags.BuildFlags` with compile flags and link flags.
        """
        return BuildFlags(self._environment.build_type, default_compile_flags=True, default_link_flags=True)

    def _add_dependency_flags(self, target):
        """Add dependencies' public and interface flags to the own and forwards their public flags.
        """
        self._build_flags.apply_public_flags(target)
        self._build_flags.forward_public_flags(target)
        self._build_flags.apply_interface_flags(target)


class StaticLibrary(Compilable):
    def __init__(self, target_description, files, dependencies=None):

        super().__init__(
            target_description=target_description,
            files=files,
            link_command=[str(target_description.environment.compiler.clang_ar), "rc"],
            output_folder=_platform.STATIC_LIBRARY_OUTPUT,
            platform_flags=_platform.PLATFORM_EXTRA_FLAGS_STATIC,
            prefix=_platform.STATIC_LIBRARY_PREFIX,
            suffix=_platform.STATIC_LIBRARY_SUFFIX,
            dependencies=dependencies,
        )

        ### Link self
        self.link_command += [
            str(buildable.object_file) for buildable in self.buildables
        ]
        self.link_command += self._build_flags.final_link_flags_list()

        ### Link dependencies
        for target in self.dependencies:
            if not target.__class__ is HeaderOnly:
                self.link_command += [
                    str(buildable.object_file) for buildable in target.buildables
                ]

    def _add_dependency_flags(self, target):
        """Add dependencies' public flags to the own and forwards their public and interface flags.

        This is done, because the dependency's interface flags will contain a header-only or static
        library's link dependencies, which cannot be applied to this static library either.
        """
        self._build_flags.apply_public_flags(target)
        self._build_flags.forward_public_flags(target)
        self._build_flags.forward_interface_flags(target)


TARGET_MAP = {
    "executable": Executable,
    "shared library": SharedLibrary,
    "static library": StaticLibrary,
    "header only": HeaderOnly,
}


class TargetDescription(_TreeEntry, _NamedLogger):
    """A hollow Target used for dependency checking.

    Before Projects actually configure targets, they first
    make sure that all dependencies etc are correctly defined.
    For this initial step, these TargetDescriptions are used.
    This is also necessary, because some of the target properties
    like the build folder, depend on the entire project structure
    and thus the two step procedure is necessary.

    TODO: Change Attributes to properties :)
    """

    def __init__(self, name: str, config: dict, parent_project):
        """Generate a TargetDescription.

        Parameters
        ----------
        name: str
            The name of this target as it will also later be named
        config : dict
            The config for this target (e.g. read from a toml)
        identifier : str
            Unique str representation of this target
        parent_project : clang_build.project.Project
            The project to which this target belongs
        """
        _NamedLogger.__init__(self, _LOGGER)

        # The "." character is used by clang-build to create unique
        # target identifiers and is therefore forbidden in naming
        if "." in name:
            error_message = self.log_message(
                f"Name contains illegal character '.': {name}"
            )
            self._logger.error(error_message)
            raise RuntimeError(error_message)

        # If no name is given, and no "output_name" is configured,
        # the output_name will be "main"
        if not name and not config.get("output_name", None):
            config["output_name"] = "main"

        # If no name is given it will be "target"
        if not name:
            name = "target"

        self.name = name
        self.config = config
        self.parent_project = parent_project

        self.only_target = False
        self.environment = self.parent_project.environment
        self._relative_directory = self.config.get("directory", "")
        self._download_directory = None

        if self.config.get("url"):
            self._download_directory = self.build_directory.parent / "external_sources"

    def __repr__(self) -> str:
        return f"clang_build.target.TargetDescription('{self.identifier}')"

    def __str__(self) -> str:
        return f"[{self.identifier}]"

    @property
    def identifier(self):
        """Return the unique identifier of this target.

        Targets are identified by their parent projects and their name as
        "[project_name.sub_project_name.target_name]".

        The default target name is "target".
        """
        return f"{self.parent_project.identifier}.{self.name}"

    @property
    def root_directory(self):
        """Return the root source directory.

        By default, the "include" and "src" directories are searched relative to
        this folder.

        The folder can be set by adding a "directory" in the config.
        If this target has external sources, it is relative to the "external_sources"
        directory, else it is relative to the parent project's directory.
        """
        if self._download_directory:
            return self._download_directory / self._relative_directory
        else:
            return self.parent_project.directory / self._relative_directory

    @property
    def build_directory(self):
        """Return the directory that serves as root build folder for the target.
        """
        if self.only_target:
            return (
                self.parent_project.build_directory
                / self.environment.build_type.name.lower()
            )
        else:
            return (
                self.parent_project.build_directory
                / self.name
                / self.environment.build_type.name.lower()
            )

    def get_sources(self):
        """Download external sources, if present, to "build_directory/external_sources".
        """
        if self._download_directory:
            url = self.config.get("url", None)
            version = self.config.get("version", None)
            _git_download_sources(url, self._download_directory, self._logger, version, self.environment.clone_recursive)


if __name__ == "__main__":
    _freeze_support()



import clang_build

"""
Sources can be generated and subsequentially a project created
"""

def get_project(directory, environment, parent=None) -> clang_build.project.Project:
    print("------------------ mylib subproject start")
    print("parent:", parent)

    project_name = "mysubproject"
    project = clang_build.project.Project(project_name, {}, directory, environment, parent=parent)

    print("project identifier: ", project.identifier)

    target_name = "mylib"
    target = clang_build.target.TargetDescription(
        target_name,
        {
            "name": target_name,
            "target_type": "static library"
        },
        project
    )

    print("target identifier: ", target.identifier)

    project.add_targets([target])

    print("------------------ mylib subproject end")
    return project


import os, sys
import unittest
import subprocess
import shutil
import logging
import io
import stat
from pathlib import Path as _Path
from multiprocessing import freeze_support

from clang_build import clang_build
from clang_build.errors import CompileError
from clang_build.errors import LinkError
from clang_build.logging_tools import TqdmHandler as TqdmHandler

def on_rm_error( func, path, exc_info):
    # path contains the path of the file that couldn't be removed
    # let's just assume that it's read-only and unlink it.
    os.chmod( path, stat.S_IWRITE )
    os.unlink( path )

def clang_build_try_except( args ):
    try:
        clang_build.build(clang_build.parse_args(args))
    except CompileError as compile_error:
        logger = logging.getLogger('clang_build')
        logger.error('Compilation was unsuccessful:')
        for target, errors in compile_error.error_dict.items():
            printout = f'Target [{target}] did not compile. Errors:\n'
            printout += ' '.join(errors)
            logger.error(printout)
    except LinkError as link_error:
        logger = logging.getLogger('clang_build')
        logger.error('Linking was unsuccessful:')
        for target, errors in link_error.error_dict.items():
            printout = f'Target [{target}] did not link. Errors:\n{errors}'
            logger.error(printout)

class TestClangBuild(unittest.TestCase):
    def test_circular_dependency(self):
        with self.assertRaisesRegex(RuntimeError, "(\[circular_project\.mylib1\] -> \[circular_project\.mylib2\] -> \[circular_project\.mylib1\])|(\[circular_project\.mylib2\] -> \[circular_project\.mylib1\] -> \[circular_project\.mylib2\])"):
            clang_build.build(clang_build.parse_args(['-d', 'test/configuration_errors/circular_dependency']))

    def test_missing_name_with_subproject(self):
        with self.assertRaisesRegex(RuntimeError, "defining a top-level project with subprojects but without a name is illegal"):
            clang_build.build(clang_build.parse_args(['-d', 'test/configuration_errors/missing_name_with_subproject']))

    def test_missing_subproject_toml(self):
        with self.assertRaisesRegex(RuntimeError, "It is not allowed to add a subproject that does not have a project file"):
            clang_build.build(clang_build.parse_args(['-d', 'test/configuration_errors/missing_subproject_toml']))

    def test_missing_subproject_name(self):
        with self.assertRaisesRegex(RuntimeError, "It is not allowed to add a subproject that does not have a name"):
            clang_build.build(clang_build.parse_args(['-d', 'test/configuration_errors/missing_subproject_name']))

    def setUp(self):
        logger = logging.getLogger('clang_build')
        logger.setLevel(logging.INFO)
        ch = TqdmHandler()
        formatter = logging.Formatter('%(message)s')
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        logger.handlers = []
        logger.addHandler(ch)

    def tearDown(self):
        if _Path('build').exists():
            shutil.rmtree('build', onerror = on_rm_error)


if __name__ == '__main__':
    freeze_support()
    unittest.main()


#!/usr/bin/env python

import setuptools
from multiprocessing import freeze_support as _freeze_support

if __name__ == '__main__':
    _freeze_support()
    setuptools.setup(
        python_requires='>=3.7',
        setup_requires=['pbr<4'],
        pbr=True)



import logging as _logging
import tqdm as _tqdm

class TqdmHandler(_logging.StreamHandler):
    def __init__(self):
        _logging.StreamHandler.__init__(self)

    def emit(self, record):
        msg = self.format(record)
        _tqdm.tqdm.write(msg)

class NamedLoggerAdapter(_logging.LoggerAdapter):
    def process(self, msg, kwargs):
        return '%s: %s' % (self.extra['tree_element'].__str__(), msg), kwargs

class NamedLogger:
    def __init__(self, logger):
        self._logger = NamedLoggerAdapter(logger, {'tree_element': self})

    def log_message(self, message: str) -> str:
        """
        """
        return f"{self.__str__()}: {message}"



"""Module containing platform specific definitions.

.. data:: PLATFORM

    The platform name as it can appear in clang-build
    TOML files.

.. data:: EXECUTABLE_PREFIX

    Prefix added to executable file names.

.. data:: EXECUTABLE_SUFFIX

    Suffix added to executable name, including extension.

.. data:: EXECUTABLE_OUTPUT

    Folder to place compiled executable in.

.. data:: SHARED_LIBRARY_PREFIX

    Prefix added to shared library file names.

.. data:: SHARED_LIBRARY_SUFFIX

    Suffix added to shared library name, including extension.

.. data:: SHARED_LIBRARY_OUTPUT

    Folder to place compiled shared library in.

.. data:: STATIC_LIBRARY_PREFIX

    Prefix added to static library file names.

.. data:: STATIC_LIBRARY_SUFFIX

    Suffix added to static library name, including extension.

.. data:: STATIC_LIBRARY_OUTPUT

    Folder to place compiled static library in.

.. data:: PLATFORM_EXTRA_FLAGS_EXECUTABLE

    Platform specific flags that are added to the compilation of
    executables.

.. data:: PLATFORM_EXTRA_FLAGS_SHARED

    Platform specific flags that are added to the compilation of
    shared libraries.

.. data:: PLATFORM_EXTRA_FLAGS_STATIC

    Platform specific flags that are added to the compilation of
    static libraries.

.. data:: PLATFORM_PYTHON_INCLUDE_PATH

    Include directory that can be added in order to use the Python
    headers.

.. data:: PLATFORM_PYTHON_LIBRARY_PATH

    Library directory that can be added in order to link to the
    Python library.

.. data:: PLATFORM_PYTHON_LIBRARY_NAME

    Python library name that can be used to link Python.

.. data:: PLATFORM_PYTHON_EXTENSION_SUFFIX

    The platform- and Python-specific suffix for binary extension
    modules.
"""

from sys import platform as _platform
from sys import version_info as _version_info
from pathlib import Path as _Path
from sysconfig import get_paths as _get_paths
from sysconfig import get_config_var as _get_config_var


if _platform == 'linux':
    # Linux
    PLATFORM = 'linux'
    EXECUTABLE_PREFIX     = ''
    EXECUTABLE_SUFFIX     = ''
    EXECUTABLE_OUTPUT     = 'bin'
    SHARED_LIBRARY_PREFIX = 'lib'
    SHARED_LIBRARY_SUFFIX = '.so'
    SHARED_LIBRARY_OUTPUT = 'lib'
    STATIC_LIBRARY_PREFIX = 'lib'
    STATIC_LIBRARY_SUFFIX = '.a'
    STATIC_LIBRARY_OUTPUT = 'lib'
    PLATFORM_EXTRA_FLAGS_EXECUTABLE = []
    PLATFORM_EXTRA_FLAGS_SHARED     = ['-fpic']
    PLATFORM_EXTRA_FLAGS_STATIC     = []
    PLATFORM_BUNDLING_LINKER_FLAGS  = ["-Wl,-rpath,$ORIGIN"]

    PLATFORM_PYTHON_INCLUDE_PATH     = _Path(_get_paths()['include'])
    PLATFORM_PYTHON_LIBRARY_PATH     = _Path(_get_paths()['data']) / "lib"
    PLATFORM_PYTHON_LIBRARY_NAME     = f"python{_version_info.major}.{_version_info.minor}"
    PLATFORM_PYTHON_EXTENSION_SUFFIX = _get_config_var('EXT_SUFFIX')

elif _platform == 'darwin':
    # OSX
    PLATFORM = 'osx'
    EXECUTABLE_PREFIX     = ''
    EXECUTABLE_SUFFIX     = ''
    EXECUTABLE_OUTPUT     = 'bin'
    SHARED_LIBRARY_PREFIX = 'lib'
    SHARED_LIBRARY_SUFFIX = '.dylib'
    SHARED_LIBRARY_OUTPUT = 'lib'
    STATIC_LIBRARY_PREFIX = 'lib'
    STATIC_LIBRARY_SUFFIX = '.a'
    STATIC_LIBRARY_OUTPUT = 'lib'
    PLATFORM_EXTRA_FLAGS_EXECUTABLE = []
    PLATFORM_EXTRA_FLAGS_SHARED     = []
    PLATFORM_EXTRA_FLAGS_STATIC     = []
    PLATFORM_BUNDLING_LINKER_FLAGS  = ["-Wl,-rpath,@executable_path"]

    PLATFORM_PYTHON_INCLUDE_PATH     = _Path(_get_paths()['include'])
    PLATFORM_PYTHON_LIBRARY_PATH     = _Path(_get_paths()['data']) / "lib"
    PLATFORM_PYTHON_LIBRARY_NAME     = f"python{_version_info.major}.{_version_info.minor}"
    PLATFORM_PYTHON_EXTENSION_SUFFIX = _get_config_var('EXT_SUFFIX')

elif _platform == 'win32':
    # Windows
    PLATFORM = 'windows'
    EXECUTABLE_PREFIX     = ''
    EXECUTABLE_SUFFIX     = '.exe'
    EXECUTABLE_OUTPUT     = 'bin'
    SHARED_LIBRARY_PREFIX = ''
    SHARED_LIBRARY_SUFFIX = '.dll'
    SHARED_LIBRARY_OUTPUT = 'bin'
    STATIC_LIBRARY_PREFIX = ''
    STATIC_LIBRARY_SUFFIX = '.lib'
    STATIC_LIBRARY_OUTPUT = 'lib'
    PLATFORM_EXTRA_FLAGS_EXECUTABLE = ['-Xclang', '-flto-visibility-public-std']
    PLATFORM_EXTRA_FLAGS_SHARED     = ['-Xclang', '-flto-visibility-public-std']
    PLATFORM_EXTRA_FLAGS_STATIC     = ['-Xclang', '-flto-visibility-public-std']
    PLATFORM_BUNDLING_LINKER_FLAGS  = []

    PLATFORM_PYTHON_INCLUDE_PATH     = _Path(_get_paths()['include'])
    PLATFORM_PYTHON_LIBRARY_PATH     = _Path(_get_paths()['data']) / "libs"
    PLATFORM_PYTHON_LIBRARY_NAME     = f"python{_version_info.major}{_version_info.minor}"
    PLATFORM_PYTHON_EXTENSION_SUFFIX = _get_config_var('EXT_SUFFIX')

else:
    raise RuntimeError('Platform ' + _platform + 'is currently not supported.')


from glob import iglob as _iglob
from pathlib import Path as _Path

from . import platform as _platform

def _get_header_files_in_folders(folders, exclude_patterns=[], recursive=True):
    delimiter = '/**/' if recursive else '/*'
    patterns  = [str(folder) + delimiter + ext for ext in ('*.hpp', '*.hxx', '*.h') for folder in folders]
    return _get_files_in_patterns(patterns)

def _get_source_files_in_folders(folders, exclude_patterns=[], recursive=True):
    delimiter = '/**/' if recursive else '/*'
    patterns  = [str(folder) + delimiter + ext for ext in ('*.cpp', '*.cxx', '*.c') for folder in folders]
    return _get_files_in_patterns(patterns)

def _get_files_in_patterns(patterns, exclude_patterns=[], recursive=True):
    included = [_Path(f) for pattern in patterns         for f in _iglob(str(pattern), recursive=recursive) if _Path(f).is_file()]
    excluded = [_Path(f) for pattern in exclude_patterns for f in _iglob(str(pattern), recursive=recursive) if _Path(f).is_file()]
    return list(f.resolve() for f in (set(included) - set(excluded)))

def get_sources_and_headers(target_name, target_options, target_root_directory, target_build_directory):
    output = {'headers': [], 'include_directories': [], 'include_directories_public': [], 'sourcefiles': []}

    # TODO: maybe the output should also include the root dir, build dir and potentially download dir?
    # TODO: should warn when a specified directory does not exist!

    # Options for include directories
    include_options = []
    include_options += target_options.get('include_directories', [])

    include_options += target_options.get(_platform.PLATFORM, {}).get('include_directories', [])

    exclude_options = []
    exclude_options += target_options.get('headers_exclude', [])

    exclude_options += target_options.get(_platform.PLATFORM, {}).get('headers_exclude', [])

    include_patterns = list(set(target_root_directory.joinpath(path) for path in include_options))
    exclude_patterns = list(set(target_root_directory.joinpath(path) for path in exclude_options))

    # Find header files
    if include_patterns:
        output['include_directories'] = include_patterns
        output['headers'] += _get_header_files_in_folders(output['include_directories'], exclude_patterns=exclude_patterns, recursive=True)
    else:
        output['include_directories'] += [target_root_directory.joinpath(''), target_root_directory.joinpath('include'),
                                        target_root_directory.joinpath(target_name)]
        output['headers'] += _get_header_files_in_folders(output['include_directories'], exclude_patterns=exclude_patterns, recursive=False)

    # Options for public include directories, exclude patterns are the same
    include_options_public = []
    include_options_public += target_options.get('include_directories_public', [])

    include_options_public += target_options.get(_platform.PLATFORM, {}).get('include_directories_public', [])

    include_patterns = list(dict.fromkeys(target_root_directory.joinpath(path) for path in include_options_public))
    exclude_patterns = list(dict.fromkeys(target_root_directory.joinpath(path) for path in exclude_options))

    # Find header files
    if include_patterns:
        output['include_directories_public'] = include_patterns
        output['headers'] += _get_header_files_in_folders(output['include_directories_public'], exclude_patterns=exclude_patterns, recursive=True)
    else:
        output['include_directories_public'] += [target_root_directory.joinpath(''), target_root_directory.joinpath('include'),
                                                target_root_directory.joinpath(target_name, 'include')]
        output['headers'] += _get_header_files_in_folders(output['include_directories_public'], exclude_patterns=exclude_patterns, recursive=False)

    # Keep only include directories which exist
    output['include_directories'] = [directory for directory in output['include_directories'] if directory.exists()]
    output['include_directories_public'] = [directory for directory in output['include_directories_public'] if directory.exists()]

    # Options for sources
    sources_options = []
    sources_options += target_options.get('sources', [])

    sources_options += target_options.get(_platform.PLATFORM, {}).get('sources', [])

    exclude_options = []
    exclude_options += target_options.get('sources_exclude', [])

    exclude_options += target_options.get(_platform.PLATFORM, {}).get('sources_exclude', [])

    sources_patterns = list(dict.fromkeys(target_root_directory.joinpath(path) for path in sources_options))
    exclude_patterns = list(dict.fromkeys(target_root_directory.joinpath(path) for path in exclude_options))

    # Find source files from patterns (recursively)
    if sources_patterns:
        output['sourcefiles'] += _get_files_in_patterns(sources_patterns, exclude_patterns=exclude_patterns, recursive=True)
    # Else search source files in folder with same name as target and src folder (recursively)
    else:
        output['sourcefiles'] += _get_source_files_in_folders([target_root_directory.joinpath(target_name), target_root_directory.joinpath('src')], exclude_patterns=exclude_patterns, recursive=True)

    # Search the root folder as last resort (non-recursively)
    if not output['sourcefiles']:
        output['sourcefiles'] += _get_source_files_in_folders([target_root_directory], exclude_patterns=exclude_patterns, recursive=False)

    # Fill return dict
    output['include_directories']        = list(dict.fromkeys(output['include_directories'] ))
    output['include_directories_public'] = list(dict.fromkeys(output['include_directories_public'] ))
    output['headers']                    = list(dict.fromkeys(output['headers'] ))
    output['sourcefiles']                = list(dict.fromkeys(output['sourcefiles'] ))

    return output